#include "fuse4bsd.h"
#include "fuse_kernel.h"

/* Mapping versions to features */

#define FUSE_KERNELABI_GEQ(maj, min)	\
(FUSE_KERNEL_VERSION > (maj) || (FUSE_KERNEL_VERSION == (maj) && FUSE_KERNEL_MINOR_VERSION >= (min)))

#ifndef USE_OLD_CLONEHANDLER_API
#if __FreeBSD_version < 600034 || ( __FreeBSD_version >= 700000 && __FreeBSD_version < 700002 )
#define USE_OLD_CLONEHANDLER_API 1
#endif
#endif

/*
 * XXX appearance of new FUSE operations is not always in par with version
 * numbering... At least, 7.3 is a sufficient condition for having
 * FUSE_{ACCESS,CREATE}.
 */
#if FUSE_KERNELABI_GEQ(7, 3)
#ifndef FUSE_HAS_ACCESS
#define FUSE_HAS_ACCESS 1
#endif
#ifndef FUSE_HAS_CREATE
#define FUSE_HAS_CREATE 1
#endif
#endif

/* Pseudo-general subroutines */

int get_unprivileged_proc_debug(struct thread *td);
int cr_candebug(int unprivileged_proc_debug, struct ucred *dcr, struct ucred *ucr);
int isbzero(void *buf, size_t len);

/* Fuse locking */

#ifndef DO_GIANT_MANUALLY
#define DO_GIANT_MANUALLY 1
#endif
#ifndef USE_FUSE_LOCK
#define USE_FUSE_LOCK 1
#endif

#if USE_FUSE_LOCK
#define FUSE_LOCK mtx_lock(&fuse_mtx)
#define FUSE_UNLOCK mtx_unlock(&fuse_mtx)
#else
#define FUSE_LOCK
#define FUSE_UNLOCK
#endif

#if USE_FUSE_LOCK
extern struct mtx fuse_mtx;
#endif

#define ASSERT_VOP_ELOCKED__FH(vp) ASSERT_VOP_ELOCKED((vp), "unsafe filehandle access")
#define ASSERT_VOP_LOCKED__FH(vp) ASSERT_VOP_LOCKED((vp), "unsafe filehandle access")


/*
 *
 * Data structures
 *
 */



/*
 * Basic data carrying thingy.
 * fuse_iov_adjust() is the primary public interface.
 */
struct fuse_iov {
	void *base;
	size_t allocated_size;		/* Quantity of space allocated to base */
	size_t len;     /* To keep track of size of the data pushed into base, =< len, of course */
};

#if ! FUSE_AUX
#ifndef FUSE_MAX_STORED_FREE_TICKETS
#define FUSE_MAX_STORED_FREE_TICKETS 0
#endif
#ifndef FUSE_MAX_PERMANENT_IOBUF	
#define FUSE_MAX_PERMANENT_IOBUF 0	
#endif

#define FUSE_DIMALLOC(fiov, spc1, spc2, amnt)				\
do { 									\
	fuse_iov_adjust(fiov, (sizeof(*(spc1)) + (amnt)));		\
	(spc1) = (fiov)->base;						\
	(spc2) = (char *)(fiov)->base + (sizeof(*(spc1)));		\
} while (0)

#define NFREE(addr, type)	\
do {				\
	FREE((addr), (type));	\
	addr = NULL;		\
} while (0)


#define FU_AT_LEAST(siz)	\
	MAX((siz), 128)

struct fuse_msg_node {
	struct fuse_iov msg;
	caddr_t msg_bufdata;
	unsigned long msg_bufsize;
	struct fuse_ticket *ticket;
	STAILQ_ENTRY(fuse_msg_node) fmsg_link;
};

struct fuse_callback_node;

typedef	int fuse_handler_t(struct fuse_callback_node *caliban, struct uio *uio);

struct fuse_callback_node {
	struct fuse_ticket *ticket;
	struct fuse_iov params; /* XXX currently not used */
	struct fuse_out_header *ohead;
	struct fuse_iov resp;
	caddr_t resp_bufdata;
	unsigned long resp_bufsize;
	int errno;
	struct mtx answer_mtx;
	fuse_handler_t *handler;
	TAILQ_ENTRY(fuse_callback_node) fcallb_link;
};

struct fuse_data;

/*
 * fuse_ticket, fuse_msg_node and fuse_callback_node are in fact
 * three sides of the same coin. The former one includes instances of the other two,
 * and those have pointers to the ticket containing them.
 */

struct fuse_ticket {
	uint64_t unique;
	struct fuse_data *data;
	struct fuse_msg_node msgn;
	struct fuse_callback_node callbn;
	int miscflag;
	unsigned age;
	STAILQ_ENTRY(fuse_ticket) freetickets_link;
	TAILQ_ENTRY(fuse_ticket) busytickets_link;
};

/*
 * The data representing a fuse session.
 * It contains the queues used in the communication
 * and the respective locks.
 */

struct fuse_data {
	struct mtx msg_mtx;
	STAILQ_HEAD(, fuse_msg_node) fmsg_head;

	struct mtx callback_mtx;
	TAILQ_HEAD(, fuse_callback_node) fcallb_head;

	struct mtx ticket_mtx;
	STAILQ_HEAD(, fuse_ticket) freetickets_head;
	TAILQ_HEAD(, fuse_ticket) busytickets_head;
	unsigned freeticket_counter;
	uint64_t ticketer;

	struct ucred *daemoncred;
	
	int dataflag;

	int mntco;
	struct mount *mp;
	struct sx slock;

	uint32_t fuse_libabi_major;
	uint32_t fuse_libabi_minor;

	unsigned max_write;
};

struct fuse_dispatcher {
	/* fields for stage 1 (and 2) */
	struct cdev *fdev;
	struct sx *slock; /* used in stage 3, too */
	struct fuse_data *data;
	/* fields for stage 2 (and 3) */
	struct fuse_ticket *tick;
	struct fuse_in_header *finh;
	void *indata;
	size_t iosize;
	uint64_t nodeid;
	/* fields for stage 3 */
	int answ_stat;
	void *answ;
};

#define FU_ANSW  0x01
#define FU_INVAL 0x02
#define FU_DIRTY 0x04

#define FDAT_KICK          0x01
#define FDAT_OPENED        0x02
#define FDAT_NOFSYNC       0x04
#define FDAT_NOFSYNCDIR    0x08
#define FDAT_NOACCESS      0x10
#define FDAT_NOCREATE      0x20

#define VTOI(vp) ((struct fuse_vnode_data *)(vp)->v_data)->nid
#define VTOI32(vp) (unsigned long long)(vp)->v_hash

/** Max number of pages that can be used in a single read request */
/* (taken from Linux Fuse) */
#define FUSE_MAX_PAGES_PER_REQ 32

#define FUSEFS_UNPRIVPROCDBG    0x01
#define FUSEFS_DAEMON_CAN_SPY   0x04
#define FUSEFS_NEGLECT_SHARES   0x08
#define FUSEFS_PRIVATE          0x10
#define FUSEFS_SECONDARY        0x20
#define FUSEFS_BUSY             0x40
#define FUSEFS_PUSH_SYMLINKS_IN 0x80
#define FUSEFS_DEFAULT_PERMISSIONS 0x100

struct fuse_mnt_data {
	struct cdev *fdev;
	struct vnode *rvp;
	int mntopts;
	struct mount *mp; 

	struct fuse_mnt_data *master;
	LIST_HEAD(, fuse_mnt_data) slaves_head;
	LIST_ENTRY(fuse_mnt_data) slaves_link;
};

struct fuse_filehandle {
	uint64_t fh_id;
	int mode;
	struct ucred *cred;
	pid_t pid;
	int useco;
	LIST_ENTRY(fuse_filehandle) fh_link;
	struct file *fp;
	int flags;
	enum fuse_opcode op;
};

struct fuse_vnode_data {
	uint64_t nid;
	uint64_t nlookup;
	//struct vnode *parent;
	/* Timeout related stuff, etc. will come here */
	int flags;
	LIST_HEAD(, fuse_filehandle) fh_head;
	int fh_counter;
#if FUSE_HAS_CREATE
	struct componentname *germcnp;
#endif
};
#define FVP_ACCESS_NOOP   0x01
#endif /* ! FUSE_AUX */

/* Debug related stuff */

#if DEBUGTOLOG
#define dprintf(args ...)  log(LOG_DEBUG, args)
#else
#define dprintf(args ...)  printf(args)
#endif

#define DEBLABEL "[fuse-debug] "
#if     _DEBUG
#ifndef _DEBUG2G
#define _DEBUG2G 1
#endif
#define DEBUG(args, ...)							\
	printf(DEBLABEL "%s:%d: " args, __func__, __LINE__, ## __VA_ARGS__)
#ifndef IGNORE_INLINE
#define IGNORE_INLINE 1
#endif
#else
#define DEBUG(args ...)
#endif

#if     _DEBUG2G
#ifndef _DEBUG3G
#define _DEBUG3G 1
#endif
#define DEBUG2G(args, ...)							\
	printf(DEBLABEL "%s:%d: " args, __func__, __LINE__, ## __VA_ARGS__)
#else
#define DEBUG2G(args ...)
#endif

#if     _DEBUG3G
#define DEBUG3G(args, ...)							\
	printf(DEBLABEL "%s:%d: " args, __func__, __LINE__, ## __VA_ARGS__)
#else
#define DEBUG3G(args ...)
#endif

#if     FMASTER
#ifndef _DEBUG_MSGING
#define _DEBUG_MSGING 1
#endif
#endif

#if     _DEBUG_MSGING
#define fuprintf(args...) \
	uprintf("[kern] " args)
#else
#define fuprintf(args...)
#endif

#if _DEBUG || _DEBUG2G || _DEBUG3G || FMASTER
extern char *fuse_opnames[];
extern int fuse_opnames_entries;

void uprettyprint(char *buf, size_t len);
void prettyprint(char *buf, size_t len);
void fprettyprint(struct fuse_iov *fiov, size_t dlen);
#endif

#if IGNORE_INLINE
#define __inline
#endif

#if _DEBUG || _DEBUG2G || _DEBUG3G
#define bp_print(bp) \
printf("b_bcount %d, b_data %p, b_error 0x%x, b_iocmd 0x%x, b_ioflags 0x%x, b_iooffset %d, b_resid %d, b_blkno %d, b_offset %d, b_flags 0x%x, b_bufsize %d, b_lblkno %d, b_vp %p, b_vp_ino %d, b_dirtyoff %d, b_dirtyend %d, b_npages %d\n", \
(int)(bp)->b_bcount, (bp)->b_data, (bp)->b_error, (bp)->b_iocmd, (bp)->b_ioflags, (int)(bp)->b_iooffset, (int)(bp)->b_resid, (int)(bp)->b_blkno, (int)(bp)->b_offset, (bp)->b_flags, (int)(bp)->b_bufsize, (int)(bp)->b_lblkno, (bp)->b_vp, VTOI((bp)->b_vp), (int)(bp)->b_dirtyoff, (int)(bp)->b_dirtyend, (int)(bp)->b_npages)
#endif
