/*-
 * Copyright (c) 2005 Jean-Sbastien Pdron
 * Copyright (c) 2005 Csaba Henk 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/mount.h>
#include <sys/uio.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>

#include "fuse4bsd.h"
#include "mntopts.h"

void	__usage_short(void);
void	usage(void);
void	helpmsg(void);
void	showversion(void);

/*
 * We don't use fsname= as of now, but if we are called by the fuse lib,
 * we will see such a thing.
 */
struct mntopt mopts[] = {
	{ "private",             0, 0x01, 1},
	{ "neglect_shares",      0, 0x02, 1},
	{ "push_symlinks_in",    0, 0x04, 1},
	{ "allow_other",         0, 0x08, 1},
	{ "default_permissions", 0, 0x10, 1},
        /* Linux specific options, we silently ignore them */
	{ "fsname=",             0, 0x00, 1},
	{ "fd=",                 0, 0x00, 1},
	{ "rootmode=",           0, 0x00, 1},
	{ "user_id=",            0, 0x00, 1},
	{ "group_id=",           0, 0x00, 1},
	{ "large_read",          0, 0x00, 1},
	{ "max_read=",           0, 0x00, 1},
	MOPT_STDOPTS,
	MOPT_END
};

const char *seekscript =
"/usr/bin/fstat /dev/fuse* | "
"/usr/bin/awk '"
"	BEGIN{ "
"		getline; "
"		if (! ($3 == \"PID\" && $4 == \"FD\" && $10 == \"NAME\")) { exit 1; }; "
"		i = 0; "
"	}; "
"	{ "
"		if ($3 == %d && $4 == %d) { "
"			i += 1; "
"			if(i > 1) { exit 1; }; "
"			printf $10; "
"		}; "
"	}; "
"	END{if (i == 0) { exit 1; }}'";

char *progname;

int
main(int argc, char *argv[])
{
	struct iovec *iov;
	char ch = '\0';
	int mntflags, iovlen, verbose = 0;
	char *dev = NULL, *dir = NULL, mntpath[MAXPATHLEN];
	char *devo = NULL, *diro = NULL;
	char ndev[128];
	int i, done = 0, no_allow_other = 0, safe_level = 0;
	int altflags = 0;
	struct mntopt *mo;
	static struct option longopts[] = {
		{"no-allow_other", no_argument, NULL, 'A'},
		{"safe", no_argument, NULL, 'S'},
		{"daemon", required_argument, NULL, 'D'},
		{"daemon_opts", required_argument, NULL, 'O'},
		{"special", required_argument, NULL, 's'},
		{"mountpath", required_argument, NULL, 'm'},
		{"version", no_argument, NULL, 'V'},
		{"help", no_argument, NULL, 'h'},
		{0,0,0,0}
	};
	int pid = 0;
	int fd = -1, fdx;
	char *ep;
	char *daemon = NULL, *daemon_opts = NULL;

	progname = argv[0];

	/*
	 * We want a parsing routine which is not sensitive to
	 * the position of args/opts; it should extract the
	 * first two args and stop at the beginning of the rest.
	 * (This makes it easier to call mount_fusefs from external
	 * utils than it is with a strict "util flags args" syntax.)
	 */

	iov = NULL;
	iovlen = 0;
	mntflags = 0;
	/* All in all, I feel it more robust this way... */
	unsetenv("POSIXLY_CORRECT");
	if (getenv("MOUNT_FUSEFS_IGNORE_UNKNOWN"))
		getmnt_silent = 1;
	if (getenv("MOUNT_FUSEFS_VERBOSE"))
		verbose = 1;

	do {
		char *inner_opt;
		int optlen;

		for (i = 0; i < 3; i++) {
			if (optind < argc && argv[optind][0] != '-') {
				if (dir) {
					done = 1;
					break;
				}
				if (dev)
					dir = argv[optind];
				else
					dev = argv[optind];
				optind++;
			}
		}
		switch(ch) {
		case 'A':
			no_allow_other = 1;
			break;
		case 'S':
			safe_level = 1;
			break;
		case 'D':
			if (daemon)
				errx(1, "daemon specified inconsistently");
			daemon = optarg;
			break;
		case 'O':
			if (daemon_opts)
				errx(1, "daemon opts specified inconsistently");
			daemon_opts = optarg;
			break;
		case 'o':
			getmntopts(optarg, mopts, &mntflags, &altflags);
			for (mo = mopts; mo->m_flag; ++mo) {
				if (altflags & mo->m_flag)
					build_iovec(&iov, &iovlen, mo->m_option, "", -1);
			}
			break;
		case 's':
			if (devo)
				errx(1, "special specified inconsistently");
			devo = optarg;
			break;
		case 'm':
			if (diro)
				errx(1, "mount path specified inconsistently");
			diro = optarg;
			break;
		case 'v': 
			verbose = 1;
			break;
		case 'h':
			helpmsg();
			break;
		case 'V':
			showversion();
			break;
		case '\0':
			break;
		case '?':
		default:
			usage();
		}
		if (done)
			break;
	} while ((ch = getopt_long(argc, argv, "AvVho:SD:O:s:m:", longopts, NULL)) != -1);

	argc -= optind;
	argv += optind;

	if (devo) {
		if (dev)
			errx(1, "special specified inconsistently");
		dev = devo;
	} else if (diro)
		errx(1, "if mountpoint is given via an option, special should also be given via an option"); 

	if (diro) {
		if (dir)
			errx(1, "mount path specified inconsistently");
		dir = diro;
	}

	if ((! dev) && argc > 0) {
		dev = *argv++;
		argc--;
	}

	if ((! dir) && argc > 0) {
		dir = *argv++;
		argc--;
	}

	if (! (dev && dir))
		errx(1, "missing special and/or mountpoint");

	/*
	 * no_allow_other is stronger than a negative of allow_other:
	 * if this is set, allow_other is blocked, period.
	 */
	if (no_allow_other) {
		for (mo = mopts; strcmp(mo->m_option, "allow_other"); ++mo);
		if (altflags & mo->m_flag)
			errx(1, "\"allow_other\" usage is banned by respective option");
	}

	if (getenv("MOUNT_FUSEFS_SAFE"))
		safe_level = 1;

	if (safe_level > 0 && (argc > 0 || daemon || daemon_opts))
		errx(1, "safe mode, spawning daemon not allowed");

	if ((argc > 0 && (daemon || daemon_opts)) || (daemon_opts && ! daemon))
		errx(1, "daemon specified inconsistently");

	/*
	 * Resolve the mountpoint with realpath(3) and remove unnecessary
	 * slashes from the devicename if there are any.
	 */
	(void)checkpath(dir, mntpath);
	(void)rmslashes(dev, dev);

	if (strcmp(dev, "auto") == 0)
		dev = "/dev/fuse";

	if (strcmp(dev, "/dev/fuse") == 0) {
		if (! (argc > 0 || daemon)) {
			fprintf(stderr, "Please also specify the fuse daemon to run when mounting via the multiplexer!\n");
			usage();
		}
		if ((fd = open(dev, O_RDWR)) < 0)
			err(1, "failed to open fuse device");
	} else {
		fdx = strtol(dev, &ep, 10);
		if (*ep == '\0')
			fd = fdx;
	}

	/* Identifying device */
	if (fd >= 0) {
		int rv;
		char *ssc;
		FILE *sf;

		asprintf(&ssc, seekscript, getpid(), fd);
		sf = popen(ssc, "r");
		if (! sf)
			err(1, "failed to run seekscript");
		fgets(ndev, sizeof(ndev), sf);
		rv = pclose(sf);
		if (rv)
			errx(1, "seekscript failed");

		dev = ndev;
	}

	if (argc > 0 || daemon) {
		char *fds;

		if (fd < 0 && (fd = open(dev, O_RDWR)) < 0)
			err(1, "failed to open fuse device");
	
		asprintf(&fds, "%d", fd);
		setenv("FUSE_DEV_FD", fds, 1);
		setenv("FUSE_NO_MOUNT", "1", 1);

		if (daemon) {
			char *bgdaemon;
			int len;

			if (! daemon_opts)
				daemon_opts = "";

			len =  strlen(daemon) + 1 + strlen(daemon_opts) + 2 + 1;
			bgdaemon = calloc(1, len);

			if (! bgdaemon)
				err(1, "failed to allocate memory");

			strlcpy(bgdaemon, daemon, len);
			strlcat(bgdaemon, " ", len);
			strlcat(bgdaemon, daemon_opts, len);
			strlcat(bgdaemon, " &", len);

			if (system(bgdaemon))
				err(1, "failed to call fuse daemon");
		} else {
			if ((pid = fork()) < 0)
				err(1, "failed to fork for fuse daemon");

			if (pid == 0) {
				execvp(argv[0], argv);
				err(1, "failed to exec fuse daemon");
			}
		}
	}

	if (fd >= 0 && close(fd) < 0) {
		if (pid)
			kill(pid, SIGKILL);
		err(1, "failed to close fuse device");
	}

	/* Prepare the options vector for nmount(). build_iovec() is declared
	 * in mntopts.h. */
	build_iovec(&iov, &iovlen, "fstype", "fusefs", -1);
	build_iovec(&iov, &iovlen, "fspath", mntpath, -1);
	build_iovec(&iov, &iovlen, "from", dev, -1);

	if (verbose)
		fprintf(stderr, "mounting fuse daemon on device %s\n", dev);

	if (nmount(iov, iovlen, mntflags) < 0)
		err(EX_OSERR, "%s on %s", dev, mntpath);

	exit(0);
}

void
__usage_short(void) {
	fprintf(stderr,
	    "usage:\n%s [-A|-S|-v|-V|-h|-D daemon|-O args|-s special|-m node|-o option...] special node [daemon args...]\n\n",
	    basename(progname));
}

void
usage(void)
{
	struct mntopt *mo;

	__usage_short();

	fprintf(stderr, "known options:\n");
	for (mo = mopts; mo->m_flag; ++mo)
		fprintf(stderr, "\t%s\n", mo->m_option);

	fprintf(stderr, "\n(use -h for a detailed description of these options)\n");
	exit(EX_USAGE);
}

void
helpmsg(void)
{
	if (! getenv("MOUNT_FUSEFS_CALL_BY_LIB")) {
		__usage_short();
		fprintf(stderr, "description of options:\n");
	}

	/*
	 * The main use case of this function is giving info embedded into general
	 * FUSE lib help output. Therefore the style and the content of the output
	 * tries to fit into there as much as possible.
	 */
	fprintf(stderr,
	        "    -o allow_other         allow access to other users\n"
	        /* "    -o nonempty            allow mounts over non-empty file/dir\n" */
	        "    -o default_permissions enable permission checking by kernel\n"
		/*
	        "    -o fsname=NAME         set filesystem name\n"
	        "    -o large_read          issue large read requests (2.4 only)\n"
	        "    -o max_read=N          set maximum size of read requests\n"
		 */
	        "    -o private             deny secondary mounting of the filesystem\n"
	        "    -o neglect_shares      don't report EBUSY when unmount attempted\n"
	        "                           in presence of secondary mounts\n" 
	        "    -o push_symlinks_in    prefix absolute symlinks with mountpoint\n"
	        );
	exit(EX_USAGE);
}

void
showversion(void)
{
	puts("mount_fusefs [fuse4bsd] version: " FUSE4BSD_VERSION);
	exit(EX_USAGE);
}
